/***************************************************************************
               uconfigdb.cpp  -  http://uconfig.sourceforge.net
               ------------------------------------------------
    begin                : Mon Sep 11 2000
    copyright            : (C) 2000 by Kir Kostchenko
    email                : kir@users.sourceforge.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *  This program is free software; you can redistribute it and/or modify   *
 *  it under the terms of the GNU Lesser General Public License as         *
 *  published by the Free Software Foundation; either version 2.1 of the   *
 *  License, or (at your option) any later version.                        *
 *                                                                         *
 ***************************************************************************/

#include <qstack.h>
#include <qintdict.h>
#include <qfile.h>
#include <qpoint.h>
#include <qsize.h>
#include <qrect.h>
#include <qcolor.h>
#include <qdir.h>
#include <qfont.h>

// Os depended (O_EXCL open). Why QFile does not support this!?
#include <fcntl.h>

//extern int open(const char *pathname, int flags, ... );
//extern int creat(const char *pathname, mode_t mode, ... );                                          
extern "C" int close(int fd);

#ifndef O_BINARY
#define O_BINARY 0
#else
#include <io.h>
#endif

#include "uconfigdb.h"

const char UConfig::dbSignature[] = "UConfig Version 1.0\n"
                                    "Use uconfig GUI to edit this file.\n"
                                    "\n"
                                    "GUI's homepage: http://uconfig.sourceforge.net.\n";


const int UCKey::knownKeyTypes = 2;  // ! be patient here 
 

void UCKey::debug()
{
  int k;
  unsigned char * inAddr   = ((unsigned char * ) &k) + sizeof( k ) - 1;
  unsigned char * outAddr = (unsigned char * )  &iKey;

  switch( t ) {
  case IntegerKey:
    // Our beauty hash.. How to do it faster?

    *(outAddr++) = UConfigPrivate::revBitTable[ *(inAddr--) ];
    *(outAddr++) = UConfigPrivate::revBitTable[ *(inAddr--) ];
    *(outAddr++) = UConfigPrivate::revBitTable[ *(inAddr--) ];
    *(outAddr++) = UConfigPrivate::revBitTable[ *(inAddr--) ];

    qWarning( "UCKey: Integer key: %d", k );
    break;
  case StringKey:
    qWarning( "UCKey: QString key: %s", (const char*) sKey->local8Bit() );
    break;
  default:
    qWarning( "UCKey: Unknown key type %d", (int)t );
  };

  return;
}


namespace UConfigPrivate
{

const unsigned char revBitTable[] = {0, 128, 64, 192, 32,
                                     160, 96, 224, 16, 144,
                                     80, 208, 48, 176, 112,
                                     240, 8, 136, 72, 200,
                                     40, 168, 104, 232, 24,
                                     152, 88, 216, 56, 184,
                                     120, 248, 4, 132, 68,
                                     196, 36, 164, 100, 228,
                                     20, 148, 84, 212, 52,
                                     180, 116, 244, 12, 140,
                                     76, 204, 44, 172, 108,
                                     236, 28, 156, 92, 220,
                                     60, 188, 124, 252, 2,
                                     130, 66, 194, 34, 162,
                                     98, 226, 18, 146, 82,
                                     210, 50, 178, 114, 242,
                                     10, 138, 74, 202, 42,
                                     170, 106, 234, 26, 154,
                                     90, 218, 58, 186, 122,
                                     250, 6, 134, 70, 198,
                                     38, 166, 102, 230, 22,
                                     150, 86, 214, 54, 182,
                                     118, 246, 14, 142, 78,
                                     206, 46, 174, 110, 238,
                                     30, 158, 94, 222, 62,
                                     190, 126, 254, 1, 129,
                                     65, 193, 33, 161, 97,
                                     225, 17, 145, 81, 209,
                                     49, 177, 113, 241, 9,
                                     137, 73, 201, 41, 169,
                                     105, 233, 25, 153, 89,
                                     217, 57, 185, 121, 249,
                                     5, 133, 69, 197, 37,
                                     165, 101, 229, 21, 149,
                                     85, 213, 53, 181, 117,
                                     245, 13, 141, 77, 205,
                                     45, 173, 109, 237, 29,
                                     157, 93, 221, 61, 189,
                                     125, 253, 3, 131, 67,
                                     195, 35, 163, 99, 227,
                                     19, 147, 83, 211, 51,
                                     179, 115, 243, 11, 139,
                                     75, 203, 43, 171, 107,
                                     235, 27, 155, 91, 219,
                                     59, 187, 123, 251, 7,
                                     135, 71, 199, 39, 167,
                                     103, 231, 23, 151, 87,
                                     215, 55, 183, 119, 247,
                                     15, 143, 79, 207, 47,
                                     175, 111, 239, 31, 159,
                                     95, 223, 63, 191, 127,
                                     255 };

  // Cache<..>::del return type:
enum DeleteResult {Deleted, TreeEmpty, Error };



class Writable {
protected:
  Q_UINT16 lockCount; // No more 65535 iterators, I think..
  Q_UINT8 dirty;
  Q_UINT8 rNeed;
  FAddr pos;

  Writable( const FAddr p );

public:
  virtual ~Writable();

  #if __UCDEBUG > 0
  void debug()
  {
    qWarning( "Writable: lockCount %d, dirty: %d, rNeed: %d, pos %d",
	      (int)lockCount, (int)dirty, (int)rNeed, (int)pos );
    return;
  }
  #endif

  inline bool isDirty() const { return dirty; }
  inline void resetDirty() 
  {
    #ifdef __UCDEBUG
      assert( !lockCount );
    #endif
    dirty = rNeed = 0;
  return;
  }

  inline int isLocked() const { return lockCount; }
  inline bool isNotLocked() const { return !lockCount; }

  void lock();
  void unLock();

  void flush( QDataStream& ds );
  inline bool needToRealloc() { return rNeed; }
protected:


  inline FAddr getPos() const { return pos; }
  void resetPos() { rNeed = true; }
private:
  void reAlloc( const FAddr p, QDataStream& ds );
  void read ( QDataStream& ds );
  void write( QDataStream& ds );

  virtual void load( QDataStream& ds ) = 0;
  virtual void save( QDataStream& ds ) = 0;

  void assignPos( const FAddr p ) { pos = p; }

  friend class Cache< Node >;
  friend class Cache< WEntry >;
};


inline Writable::Writable( const FAddr p )
{
   #ifdef __UCDEBUG
      assert( p );
   #endif
   
   lockCount = 0;
   pos = p;
   rNeed = false;
   
   dirty = true;
}
   

inline Writable::~Writable()
{
  #ifdef __UCDEBUG
    assert( !dirty );
    assert( !lockCount );
    assert( !rNeed );
  #endif
}


inline void Writable::reAlloc(
  const FAddr p,
  QDataStream &ds )
{
  #ifdef __UCDEBUG
    assert( pos );
  #endif

  pos = p;
  rNeed = false;
  write( ds );

  return;
}


inline void Writable::lock()
{
  lockCount++;

  #ifdef __UCDEBUG
    assert( lockCount );  // Not over 255
  #endif

return;
}


inline void Writable::unLock()
{
  #ifdef __UCDEBUG
    assert( lockCount );
  #endif

  lockCount--;

return;
}

inline void Writable::flush( QDataStream & ds )
{
  if( dirty )
    write( ds );

  return;
}


inline void checkDeviceStatus( const QIODevice *d, const char *errorHeader )
{
  switch( d->status() ) {
    case IO_Ok:
      break;
    case IO_ReadError:
      qFatal( "%s: Could not read from the file.", errorHeader );
    case IO_WriteError:
      qFatal( "%s:  Could not write to the device.", errorHeader );
    case IO_FatalError:
      qFatal( "%s: A fatal unrecoverable error occurred.", errorHeader );
    case IO_OpenError:
      qFatal( "%s: Could not open the device.", errorHeader );
    case IO_AbortError:
      qFatal( "%s: The operation was unexpectedly aborted.", errorHeader );
    case IO_TimeOutError:
      qFatal( "%s: The operation timed out.", errorHeader );
    default:
      qFatal( "%s: Unknown file status. Read failed.", errorHeader );
  }
return;
}


inline void Writable::write( QDataStream & ds )
{
  if ( !pos || !ds.device()->at( pos ) )
    qFatal( "UConfig::Writable::write: can not set pos. Pos: %d", pos );

  save( ds );

  checkDeviceStatus( ds.device(),
                     "UConfig::Writable::write" );
  dirty = false;
return;
}


inline void Writable::read ( QDataStream & ds )
{
  if ( !pos || !ds.device()->at( pos ) )
    qFatal( "UConfig::Writable::read can not set pos. Pos: %d", pos );

  load( ds );

  checkDeviceStatus( ds.device(),
                     "UConfig::Writable::read" );
  dirty = false;
return;
}


class Header {
protected:
  FAddr left;
  FAddr right;
public:

  Header() { left = right = 0; }
  virtual ~Header() { }

  FAddr getLeft()  const { return left; }
  FAddr getRight() const { return right; }

  #if __UCDEBUG > 0
  void debug()
  {
    qWarning( "Header: left - %d, right - %d", left, right );
    return;
  }
  #endif
        // Return qChecksum of readed/written data.
protected:
  Q_INT16 checksum() const;
  void load( QDataStream& ds );
  void save( QDataStream& ds );
};


inline Q_INT16 Header::checksum() const
{
return qChecksum( (char *)&left, sizeof(left) ) +
       qChecksum( (char *)&right, sizeof(right) );
}


inline void Header::load( QDataStream& ds )
{
  ds >> left >> right;

return;
}


inline void Header::save( QDataStream& ds )
{
  ds << left << right;

return;
}


class WHeader : public UCKey,
                public Header,
                public Writable {
public:
    // We will read it.
  inline WHeader( const KeyType kt,
                  const FAddr pos ) : UCKey(kt),
                                      Writable( pos ) { }

    // New header.
  inline WHeader( const UCKey &k,
                  const FAddr pos,
                  const ParentType pt = RootChild,
                  WHeader *ph = 0 ) : UCKey( k ),
                                      Writable( pos )
  {
    switch( pt ) {
      case RootChild:
        break;
      case LeftChild:
        if( ph )
          ph->setLeft( this );
        break;
      case RightChild:
        if( ph )
          ph->setRight( this );
        break;
      default:
        #ifdef __UCDEBUG
          assert( 0 );
        #endif

        ;
    }
  }


  ~WHeader() { }

  inline FAddr getPos() const { return Writable::getPos(); }


  inline void setLeft ( const WHeader *n )
  {
    left = n->pos;
    dirty = true;

  return;
  }


  inline void setRight( const WHeader *n )
  {
    right = n->pos;
    dirty = true;

  return;
  }


  inline void setLeft ( const FAddr l )
  {
    left = l;
    dirty = true;

  return;
  }


  inline void setRight( const FAddr r )
  {
    right = r;
    dirty = true;

  return;
  }

  void debug()
  {
    qWarning( "Header debug: " );
    UCKey::debug();
    Header::debug();
    Writable::debug();
    return;
  }

protected:

  inline void load( QDataStream& ds )
  {
    Q_UINT16 readedChecksum;

    UCKey::load( ds );
    Header::load( ds );

    ds >> readedChecksum;

    Q_UINT16 checksum = UCKey::checksum() + Header::checksum();

    if( readedChecksum != checksum )
      qFatal( "void UConfig::WHeader::"
              "load( QDataStream& ds ): checksum fault" );
  return;
  }


  inline void save( QDataStream& ds )
  {
    UCKey::save( ds );
    Header::save( ds );

    Q_UINT16 checksum = UCKey::checksum() + Header::checksum();
    ds << checksum;

  return;
  }
};

class WHeaderList : public QList<WHeader> {}; // Opening template.

class Record : public Writable {
  WHeader* header;
public:
  Record( WHeader *h ) : Writable( (FAddr)-1 )  // It will be
                                                // corrected later.
  {
    header = h;
    header->lock();
  }

  ~Record()
  {
    header->unLock();

    // Header could be safely locked by iterator
    //    #ifdef __UCDEBUG
    //      assert( !header->isLocked() );
    //    #endif
  }

  WHeader* getHeader() const { return header; }
};

inline Q_UINT32 variantSize (QVariant::Type t)
{
  switch (t)
  {
    // Are those constants inside QT?
    case QVariant::Int    : return 4;
    case QVariant::Double : return 8;
    case QVariant::UInt   : return 4;
    case QVariant::Bool   : return 1;
    case QVariant::Rect   : return 16;
    case QVariant::Size   : return 8;
    case QVariant::Color  : return 4;
    case QVariant::Point  : return 8;
    default:
      return 0;
  }
}

inline Q_UINT32 variantSize( const QVariant &v )
{
  Q_UINT32 size = variantSize( v.type() );
  if( !size ) {
    QByteArray dummy;
    QDataStream stream (dummy, IO_WriteOnly);
    stream << v;
    return dummy.size();
  }
  else
    return size;
}


class WEntry : public Record {
  Q_UINT32 space;
  QVariant data;
public:


  inline WEntry( WHeader *h ) : Record( h ) { space = 0;}

  inline ~WEntry() {}

  inline bool update( const QVariant &from )
  {
      // We newer going here with invalid QVariant.
    // if (data.type() == from.type() && data == from)
    //  return true;

    // it has changed, so it has to be written back
    dirty = true;

    Q_UINT32 neededSpace = variantSize (from);

    if (neededSpace > space)
    {
        // We don't need to reset Pos on first init.
      if (space)
        resetPos();

      space = neededSpace;
    }

    data = from;
    return true;
  }

protected:


  inline void load( QDataStream& ds )
  {
    ds >> data;

    space = variantSize( data );
  }


  inline void save( QDataStream& ds )
  {
    ds << data;
  }

  friend class UConfig;
  friend class UCEntriesIterator;
};


  // FIXME it can be optimized.
class Node : public Record {
  FAddr parentNode;

    // *2 for space to child nodes.
  FAddr child[ UCKey::knownKeyTypes * 2 ];

  Q_UINT8 space;  // Space after current node
  Q_UINT8 used;   // Used child pointers.
public:

  inline Node( WHeader * h,
               const Node* pn = 0,
               const int _space = 2 ) : Record( h )
  {
    if ( pn ) { // Creating.
      parentNode = pn->getHeader()->getPos();
      used = 0;
      space = _space; // By default 2 free pointers
    }
    else {
      parentNode = 0;
      used = 0;
      space = 0;
    }
      // All compillers do it by default?
    memset( child, 0, sizeof(child) );
  }

  inline ~Node() { };

  inline QString name() const { return getHeader()->getStringKey(); }

  inline FAddr getParentNode() const { return parentNode; }

  inline FAddr getSubNodes( const UCKey::KeyType t ) const { return child[t]; }
  inline FAddr getEntries ( const UCKey::KeyType t ) const
  {
    return child[ t + UCKey::knownKeyTypes ];
  }

  inline bool isInfo()
  {
    for( int i = 0; i < UCKey::knownKeyTypes * 2; i++ )
      if( child[i] )
	return true;
  return false;
  }

  inline void setParent( const Node * to )
  {

    parentNode = to->getHeader()->getPos();
    dirty = true;
  return;
  }


    // Be very patient here! Only UConfig
    // constructor call this method.
  inline void increaseSpace()
  {
    assert( !space );
    space++;
    return;
  }


  inline void setSubNodes( const Node * to )
  {
    #ifdef __UCDEBUG
      assert( to->getHeader()->getPos() );
    #endif

      // child subnodes was not allocated.
    if( ! child[ to->getHeader()->getKeyType() ] ) {
      used++;
      if( space < used ) {
        resetPos();  // we need reallocate it.
        space = used;
      }
    }

    child[ to->getHeader()->getKeyType() ] = to->getHeader()->getPos();
    dirty = true;
  return;
  }


  inline void setSubNodes( const UCKey::KeyType t, const FAddr newAddr )
  {
      int i = t;
      
      #ifdef __UCDEBUG
         assert( child[i] );
	 assert( newAddr );
	 assert( newAddr != child[i] );
      #endif

      child[i] = newAddr;
      dirty = true;
   return;
   }


  inline void setEntries( const WEntry * to )
  {
    #ifdef __UCDEBUG
      assert( to->getHeader()->getPos() );
    #endif

    int i = to->getHeader()->getKeyType() +
            UCKey::knownKeyTypes;

    if( ! child[ i ] ) {
      used++;
      if( space < used ) {
        resetPos();  // we need reallocate it.
        space = used;
      }
    }

    child[ i ] = to->getHeader()->getPos();
    dirty = true;
  return;
  }


  inline void setEntries( const UCKey::KeyType t, const FAddr newAddr )
  {
      int i = t + UCKey::knownKeyTypes;
      
      #ifdef __UCDEBUG
         assert( child[i] );
	 assert( newAddr );
	 assert( newAddr != child[i] );
      #endif

      child[i] = newAddr;
      dirty = true;
   return;
   }
      


   inline void resetEntries( const UCKey::KeyType t )
   {
      int i = t + UCKey::knownKeyTypes;
      
      #ifdef __UCDEBUG
         assert( child[i] );
      #endif

      child[i] = 0;
      used--;
      dirty = true;
   return;
   }


   inline void resetSubNodes( const UCKey::KeyType t )
   {
      int i = t;
      
      #ifdef __UCDEBUG
         assert( child[i] );
      #endif

      child[i] = 0;
      used--;
      dirty = true;
   return;
   }


protected:
  void load( QDataStream& ds )
  {
    #ifdef __UCDEBUG
      assert( Writable::getPos() != (FAddr)-1 );
      assert( !space );
    #endif

    ds >> parentNode;

    Q_UINT8 rule;

    for( space = 0,
         ds >> rule; rule != 0xFF; space++,
                                   ds >> rule ) {

      if( rule & 0x80 ) { // Bit 8 is up. subnode pointer.
        if( (rule & 0x7F) < UCKey::knownKeyTypes )
          ds >> child[ rule&0x7F ];
        else
          qFatal( "void UConfigPrivate::Node::load( QDataStream& ds ): invalid subnode key type." );

          // Here something is..
        if( child[ rule&0x7F ] )
          used++;
      }
      else {              // Entries tree pointer
        if( rule < UCKey::knownKeyTypes ) {
          Q_UINT32 p;
          ds >> p;
          if( p )
            if( !child[ rule + UCKey::knownKeyTypes ] ) {
              child[ rule + UCKey::knownKeyTypes ] = p;
              used++;
            }
            else
              qFatal( "UConfigPrivate::Node::load( QDataStream& ds ): child replacement" );
        }
        else
          qFatal( "void UConfigPrivate::Node::load( QDataStream& ds ): invalid entries tree key type." );
      }
    }
  return;
  }


  void save( QDataStream& ds )
  {
    #ifdef __UCDEBUG
      assert( Writable::getPos() != (FAddr)-1 );
      assert( Writable::getPos() );
    #endif

    ds << parentNode;
    Q_UINT8 i;
    Q_INT8 usedChild = -1;
    bool eoUsed = false;
    for( i = 0; i < space; i++ ) {

        // Run to next used child or eof.
      for( usedChild++; !eoUsed &&
                        usedChild < UCKey::knownKeyTypes*2 &&
                        !child[usedChild];
                                           usedChild++ );

      if( usedChild >= UCKey::knownKeyTypes * 2 )
        eoUsed = true;  // A bit hack ;)

        // write keytype
      if( eoUsed ) {
        ds << (Q_UINT8)0 << (Q_UINT32)0;
      }
      else {
        Q_UINT8 rule = usedChild < UCKey::knownKeyTypes ?
                       (0x80 | usedChild) :
                       (usedChild - UCKey::knownKeyTypes);

        ds << rule;

        ds << child[ usedChild ];
      }
    }

      // write end of child pointers.
    ds << (Q_UINT8)0xFF;
  return;
  }

  friend Cache< Node >;
};



template< class RecordType >
class Cache {

    // FIXME, we need not QIntDict.
  UCKey::KeyType headersKeyType;

  QIntDict< WHeader > headers;
  QIntDict< RecordType > records;

  Q_UINT32 statisticMax;
public:
  Cache( const UCKey::KeyType t )
  {
    headersKeyType = t;
    resetStatistic();

    headers.setAutoDelete( true );
    records.setAutoDelete( true );
  }


  ~Cache()
  {
    if( records.count() )
      qFatal( "UConfigPrivate::Cache::~Cache: records is not free. Not all iterators deleted?" );
    
    if( headers.count() ) {
      QIntDictIterator< WHeader > it(headers);
      while( it.current() ) {
	it.current()->debug();
	++it;
      }

      qFatal( "UConfigPrivate::Cache::~Cache: headers is not free. Not all iterators deleted?" );
    }
  }


  void resetStatistic() { statisticMax = 0; }
  Q_UINT32 statistic() { return statisticMax; }

  WHeader* safeGetHeader( const FAddr from,
                          QDataStream &ds,
                          const bool reRead = false )
  {
    if( !from )
      return 0;

    WHeader *h;
    h = headers[ from ];

    if( !h ) {  // Read it
      h = new WHeader( headersKeyType, from );
		
      h->read( ds );
		
      headers.insert( from, h );
    }
    else if ( reRead ) {
      h->flush( ds ); // Is it dirty?
      h->read( ds );
    }

  return h;
  }


  RecordType* safeGet( const FAddr from, QDataStream &ds )
  {
    if( !from )
      return 0;

    RecordType *r;
  	
    r = records[ from ];

    if( !r ) {
      WHeader *nh = safeGetHeader( from, ds, true );

      #ifdef __UCDEBUG
        assert( nh );
      #endif

      r = new RecordType( nh );
      r->assignPos( ds.device()->at() );
		
      r->read( ds );
		
      records.insert( from, r );
    }
  return r;
  }


    // FIXME - optimize it
  inline RecordType* safeGet( const WHeader * h,
                       QDataStream &ds )
  {
    return safeGet( h->getPos(), ds );
  }


  void add( RecordType *r, QDataStream& ds, Q_UINT32 *lp )
  {
    #ifdef __UCDEBUG
      assert( !headers[ r->getHeader()->getPos() ] );
      assert( !records[ r->getPos() ] );
      assert( r->getHeader()->getPos() );
      assert( r->getPos() == (Q_UINT32)-1 );
    #endif

    headers.insert( r->getHeader()->getPos(),
                    r->getHeader() );

    records.insert( r->getHeader()->getPos(), r );

    r->getHeader()->flush( ds );
    r->assignPos( ds.device()->at() );

    r->flush( ds );

    *lp = ds.device()->at();

  return;
  }


  Q_UINT32 flush( QDataStream& ds,
                  const bool free = true )
  {
  	Q_UINT32 countRecords = 0;
	
  	QIntDictIterator< RecordType > itRecord( records );

    while ( itRecord.current() ) {

      if ( itRecord.current()->isDirty() )
        countRecords++;
		
      itRecord.current()->flush( ds );

      if( free && itRecord.current()->isNotLocked() ) {

//cout << "Removing record: " << itRecord.current()->name() << endl;

        records.remove( itRecord.currentKey() );
        continue;
      }
      ++itRecord;
    }

  	QIntDictIterator< WHeader > itHeader( headers );
    while ( itHeader.current() ) {

      itHeader.current()->flush( ds );

      if( free && itHeader.current()->isNotLocked() ) {

//cout << "Removing header: " << itHeader.current()->getKey() << endl;

        headers.remove( itHeader.currentKey() );
        continue;
      }

      ++itHeader;
    }

    ds.device()->flush();

    checkDeviceStatus( ds.device(), "flush: " );

  return countRecords;
  }


  RecordType* find( const FAddr root,
		    QDataStream& ds,
		    const UCKey &k,
		    WHeader **parentHeader = 0,
		    ParentType *pt = 0 )
  {
    WHeader *h = 0;

    Q_UINT32 statisticCount = 1;

    h = safeGetHeader( root, ds );

    if( parentHeader )
      *parentHeader = 0;

    #if __UCDEBUG >= 2
      qWarning("Find - searching for '%s'",  (const char *)k.getKey().toString().local8Bit() );
    #endif

    if ( !h ) {
      if( pt )
        *pt = RootChild;
    }
    else {
      if( pt )
        *pt = FoundFromRoot;

      for( ; true ; statisticCount++ ) {

        if ( k < *h ) {

          if( parentHeader )
            *parentHeader = h;

          #if __UCDEBUG >= 2
	    qWarning("  '%s' < '%s', lock: %d, pos: %d", 
		     (const char *)k.getKey().toString().local8Bit(),
		     (const char *)h->getKey().toString().local8Bit(),
		     h->isLocked(),
		     h->getPos() );
          #endif

          if ( h->getLeft() ) {
              h = safeGetHeader( h->getLeft(), ds );
              if( pt )
                *pt = FoundFromLeft;
              continue;
          }
          else {
              if( pt )
                *pt = LeftChild;
              h = 0;
              statisticCount++;
              break;
          }
        }
        else if ( k > *h ) {

          if( parentHeader )
            *parentHeader = h;

          #if __UCDEBUG >= 2
	    qWarning("  '%s' > '%s', lock: %d, pos: %d", 
		     (const char *)k.getKey().toString().local8Bit(),
		     (const char *)h->getKey().toString().local8Bit(),
		     h->isLocked(),
		     h->getPos() );
          #endif


          if ( h->getRight() ) {
              h = safeGetHeader( h->getRight(), ds );
              if( pt )
                *pt = FoundFromRight;
              continue;
          }
          else {
              if( pt )
                *pt = RightChild;
              h = 0;
              statisticCount++;
              break;
          }
        }

        #if __UCDEBUG >= 2
	    qWarning("  '%s' == '%s', lock: %d, pos: %d", 
		     (const char *)k.getKey().toString().local8Bit(),
		     (const char *)h->getKey().toString().local8Bit(),
		     h->isLocked(),
		     h->getPos() );
        #endif

        break;
      }
    }

    if( statisticCount > statisticMax )
      statisticMax = statisticCount;


  return (h ? safeGet( h, ds ) : 0);
  }


  void reAlloc( RecordType* r,
		QDataStream& ds,
		FAddr *lastPosition )
  {
    // remove r from caches.
    #ifdef __UCDEBUG
      assert( headers.take( r->getHeader()->getPos() ) );
      assert( records.take( r->getHeader()->getPos() ) );
    #else
      headers.take( r->getHeader()->getPos() );
      records.take( r->getHeader()->getPos() );
    #endif

    r->getHeader()->reAlloc( *lastPosition, ds );
    r->reAlloc( ds.device()->at(), ds );
    *lastPosition = ds.device()->at();

    headers.insert( r->getHeader()->getPos(), r->getHeader() );
    records.insert( r->getHeader()->getPos(), r );

  return;
  }


  inline WHeader* countLeft( WHeader *from, QDataStream& ds, Q_UINT32& c )
  {
    WHeader *h = from;

      // c is 0 already (sets by caller).
    for( c=0; h->getLeft(); h = safeGetHeader(h->getLeft(), ds), c++ );
    return h;
  }


  inline WHeader* countRight( WHeader *from, QDataStream& ds, Q_UINT32& c )
  {
    WHeader *h = from;

    for( c = 0; h->getRight(); h = safeGetHeader(h->getRight(), ds), c++ );
    return h;
  }

  
  DeleteResult del( const FAddr root, 
		    QDataStream& ds, 
		    const UCKey& k,
		    FAddr& rootMovedTo )
  {
    RecordType *r;
    WHeader *h;
    WHeader *ph;
    ParentType pt;
    
    r = find( root, ds, k, &ph, &pt );

    if( !r )
      return Error; // No such entry

    if( r->isLocked() )
      return Error; // Record locked

    h = r->getHeader(); // Remove record lock from record header

      // Remove record from cache.
    r->flush( ds ); // Record could be dirty.
    records.remove( h->getPos() );

    if( h->isLocked() )
      return Error; // Header locked


    FAddr newAddr;

    if( h->getLeft() && h->getRight() ) {
      WHeader *mostLeftOnRight, *mostRightOnLeft;
      Q_UINT32 mostLeftOnRightDepth, mostRightOnLeftDepth;

      mostRightOnLeft = countRight( safeGetHeader( h->getLeft(), ds ), ds,
				    mostRightOnLeftDepth );
      mostLeftOnRight = countLeft( safeGetHeader( h->getRight(), ds ), ds,
				   mostLeftOnRightDepth );

      if( mostRightOnLeftDepth <= mostLeftOnRightDepth ) {
	newAddr = h->getLeft();
	mostRightOnLeft->setRight( h->getRight() );
      }
      else {
	newAddr = h->getRight();
	mostLeftOnRight->setLeft( h->getLeft() );
      }
    }
    else if ( (newAddr = h->getLeft()) );
    else if ( (newAddr = h->getRight()) );


    switch( pt ) {
      case UConfigPrivate::FoundFromLeft:
	ph->setLeft( newAddr );
	rootMovedTo = 0;
	break;

      case UConfigPrivate::FoundFromRight:
	ph->setRight( newAddr );
	rootMovedTo = 0;
	break;

      case UConfigPrivate::FoundFromRoot:
	if( !newAddr ) {
	  h->resetDirty();
	  headers.remove( h->getPos() );
	  return TreeEmpty;
	}
	rootMovedTo = newAddr;
	break;

      default:
	qFatal( "UConfigPrivate::Cache<..>::del(..): Parent type mismatch" );
    }

      // Remove header from cache.
    h->resetDirty();  // It could be dirty.
    headers.remove( h->getPos() );

  return Deleted;
  }
	 
};


}


QString UConfig::encode( const QString& path )
{
   QString ret;
   Q_UINT32 i; 
 
   for( i = 0; i < path.length(); i++ ) {
      if( path.at(i) == QChar('/') ||  
          path.at(i) == QChar('.') ||  
          path.at(i) == QChar('\\') )
         ret.append( QChar('\\') );
      ret.append( path.at(i) );
   }
   return ret;
}

QString UConfig::decode( const QString& path )
{
   QString ret;
   Q_UINT32 i;
   for( i = 0; i < path.length(); i++ ) {
      if( path.at(i) == '\\' ) {
         i++; // skip only one '\'
         if( i == path.length() ) break;
      }
      ret.append( QChar(path.at(i)) );
   }
   return ret;
}

UConfig::FileType UConfig::isLooksLikeConfigFile( const QString & fileName )
{
  QFile f( fileName );

  if ( f.open(IO_ReadWrite) ) {

    if ( !f.size() )
      return EmptyDB;

    QDataStream ds( &f );

    char signature[ sizeof(dbSignature) ];

    ds.readRawBytes( signature, sizeof(signature) );
    if ( !memcmp(signature, dbSignature, sizeof(signature)) )
      return PlainDB;
  }

return Unknown;
}


class UConfig::LockedDir {
  UConfigPrivate::Node *dir;
public:
  inline LockedDir( UConfigPrivate::Node *root )
  {
    dir = root;
    dir->lock();
  }

  inline ~LockedDir() { dir->unLock(); }


  inline void changeDir( UConfigPrivate::Node * to )
  {
    dir->unLock();
    dir = to;
    dir->lock();
  return;
  }

  inline operator UConfigPrivate::Node* () { return dir; }

  inline UConfigPrivate::Node* d() { return dir; }
};


int lazyOpen( const char *fn )
{
  int h = open( fn, O_BINARY | O_RDWR | O_EXCL );

  if( h == -1 ) { // Try creat, then open
    h = creat( fn, 0644 );

    if( h == -1 )
      return h;

    close( h );

    h = open( fn, O_BINARY | O_RDWR | O_EXCL );
  }
  return h;
}


UConfig::UConfig( const QString &configFileName,
                  const bool fromHomeDir )
{
  QFile *file = new QFile;

  QString fileName;

  if(fromHomeDir) // HomeDir + '/'
    fileName = QDir::homeDirPath() + QDir::separator();

  fileName += configFileName;

  file->setName( fileName );

  nodes = new UConfigPrivate::NodesCache( UCKey::StringKey );
  eCaches = new UConfigPrivate::ECache*[ UCKey::knownKeyTypes ];

  int i;
  for( i = 0; i < UCKey::knownKeyTypes; i++ )
    eCaches[i] = new UConfigPrivate::ECache( (UCKey::KeyType)i );

  int h;
  if( (h = lazyOpen( fileName.local8Bit() )) == -1 )
    if( fromHomeDir ) {

        // Let's try to create directories inside homedir.
      int sepPos;
      for( sepPos = 0;
           (sepPos = configFileName.find( QDir::separator(), sepPos )) != -1;
           sepPos++ )

        if( !QDir::home().mkdir( configFileName.left(sepPos) ) )
          qFatal( "UConfig::UConfig(..): Could not create directory %s for config file",
                  (const char *)(QDir::homeDirPath() +
                                 QDir::separator() +
                                 configFileName.left(sepPos)).local8Bit() );

        // Directories created. Try to open it again.
      if( (h = lazyOpen( fileName.local8Bit() )) == -1 )
        qFatal( "UConfig::UConfig(..): Could not open %s.", 
		(const char *)fileName.local8Bit() );
    }
    else
      qFatal( "UConfig::UConfig(..): Could not open %s.", 
	      (const char *)fileName.local8Bit() );


  if( !file->open( IO_ReadWrite, h ) )
    qFatal( "UConfig::UConfig(..): file->open( IO_ReadWrite, h ) failed." );
  file->resetStatus();

  ds.setDevice( file );

  file->at( 0 );

  if ( file->size() ) { // File exists
    char signature[ sizeof(dbSignature) ];
    ds.readRawBytes( signature, sizeof(signature) );

    if ( memcmp(signature, dbSignature, sizeof(signature)) )
      qFatal( "UConfig::UConfig(..): Database %s signature wrong.", (const char *)fileName.local8Bit() );

    else {  // Read root and make it current.
            // Set currrent path to '/'
      ds >> rootPosition;
      file->at( rootPosition );
      curDir = new LockedDir( nodes->safeGet( rootPosition, ds ) );
      lastPosition = file->size();

      if( !curDir )
      qFatal( "UConfig::UConfig(..): can not read root node" );				
				
    }
  }
  else {  // New database
    ds.writeRawBytes( dbSignature, sizeof(dbSignature) );

      // Pointer to root entry.
    ds << (Q_UINT32)( file->at() + sizeof( Q_UINT32) );
			
    rootPosition = file->at();

    curDir = new LockedDir(
               new UConfigPrivate::Node(
                 new UConfigPrivate::WHeader(QString::null, rootPosition)
                                       )
                          );

      // By default root does not have any entries.
      // Only subnodes.
    curDir->d()->increaseSpace();
    nodes->add( *curDir, ds, &lastPosition );
  }
}


UConfig::~UConfig()
{
  // unlock curDir record.
  delete curDir;
	
  // detach iterators.
  for( iterators.first();
       iterators.current();
       iterators.next()
       )
    iterators.current()->detach();
	
	
  // flush and delete all nodes from cache.	
  nodes->flush( ds );
  delete nodes;
	
    // flush and delete all entries caches.
  int i;
  for( i = 0; i < UCKey::knownKeyTypes; i++ ) {
    eCaches[i]->flush( ds );
    delete eCaches[i];
  }

  delete eCaches;

	
  QIODevice * f = ds.device();
  f->close();
  ds.unsetDevice();
  delete f;
}


QString UConfig::path( UConfigPrivate::Node *p)
{
   const int maxRecurse = 1000;
   int rec = 0;

   QString ps;
   if ( p ) {
      QStack< UConfigPrivate::Node > stack;

      while( p ) {
         stack.push( p );
         p->lock();
			
         p = nodes->safeGet( p->getParentNode(), ds );

         if ( ++rec > maxRecurse )
            qFatal( "Recurse too long: QString UConfig::path()" );
      }

      if ( rec > 1 )
         stack.pop()->unLock();	// For skip first '/'

      while( !stack.isEmpty() ) {
         p = stack.pop();
//           cerr << endl << "popped group " << (const char*)p->name().local8Bit()
//                << ", encoded to " << (const char*)encode(p->name()).local8Bit() << endl;
         ps += "/" + encode(p->name());
//           cerr << "Growing path is " << (const char*)ps.local8Bit() << endl;
         p->unLock();
      }
   }
   else
      qFatal( "UConfig::path( UConfigPrivate::Node* p) p is null" );

//     cerr << "***returning path is " << (const char*)ps.local8Bit() << endl;
   return ps;
}


QString UConfig::path() { return path( *curDir ); }


inline UConfigPrivate::Node*
UConfig::appendGroup( const UCKey& k,
                      const int space,
                      UConfigPrivate::WHeader *ph,
                      UConfigPrivate::ParentType pt )
{
   // Append group.
   UConfigPrivate::Node* newNode;
   newNode =  new UConfigPrivate::Node(
      new UConfigPrivate::WHeader( k,
                                   lastPosition,
                                   pt,
                                   ph ),
      *curDir,
      space );

   nodes->add( newNode, ds, &lastPosition );


   if ( pt == UConfigPrivate::RootChild ) { // First child
      curDir->d()->setSubNodes( newNode );

      // Ooh..
      if( curDir->d()->needToRealloc() )
         reAllocCurDir();
   }


   if( ph )  // Crash protection. Reduces speed.
      ph->flush( ds );

return newNode;
}


bool UConfig::setPath( const QString &path,
                       const bool createNew,
		       const int space )
{
   if ( !curDir )
      return false;

   QString simplePath = path;

   QString dir;

   if ( simplePath[0] == '/' ) {
      if ( curDir->d()->getParentNode() )
         curDir->changeDir( nodes->safeGet( rootPosition, ds ) );
      simplePath = simplePath.mid( 1 );
   }

   while ( !simplePath.isEmpty() ) {
      // Skip empty '/'
      while ( !simplePath.isEmpty() && simplePath[0] == '/' )
         simplePath = simplePath.mid( 1 );
      
      for( dir = QString::null;
           !simplePath.isEmpty() && simplePath[0] != '/';
           simplePath = simplePath.mid( 1 ) ) {
         if( simplePath[0] == '\\' ) {        // ex: make "\/" sound like "/" to be able
            simplePath = simplePath.mid( 1 ); // to use special characters in config
            dir += QString(QChar('\\')) + simplePath[0];
            continue;
         }
         dir+=simplePath[0];
      }

      if ( dir == "." )	// Do nothing
         continue;

      if ( dir == ".." ) // Go to parent.
         if ( curDir->d()->getParentNode() ) {
            curDir->changeDir( nodes->safeGet( curDir->d()->getParentNode(), ds ) );
            continue;
         }
         else {
            return false;
         }

      // It's a dir.
      if( !dir.isEmpty() ) {
         UConfigPrivate::Node *found;
         UConfigPrivate::WHeader *ph;
         UConfigPrivate::ParentType pt;

         UCKey k( decode(dir) );

         found = nodes->find( curDir->d()->getSubNodes(UCKey::StringKey),
                              ds,
                              k,
                              &ph,
                              &pt );
         
         if( !found )
            if( createNew )
               found = appendGroup( k, space, ph, pt );
            else
               return false; // FIXME (unlock found)
		  
         curDir->changeDir( found );
      }
   }
return true;
}


Q_UINT32 UConfig::flush( bool free )
{
   Q_UINT32 count = nodes->flush( ds, free );	

   // flush and delete all entries caches.
   int i;
   for( i = 0; i < UCKey::knownKeyTypes; i++ )
      count += eCaches[i]->flush( ds );

return count;
}


bool UConfig::readEntry (const UCKey &k, bool &v, bool d)
{
   QVariant e = internalReadEntry (k);
   if (e.canCast (QVariant::Bool)) {
     v = e.toBool();
     return true;
   } else {
     v = d;
     return false;
   }
}


bool UConfig::readEntry (const UCKey &k, int &v, int d)
{
   QVariant e = internalReadEntry (k);
   if (e.canCast (QVariant::Int)) {
     v = e.toInt();
     return true;
   } else {
     v = d;
     return false;
   }
}


bool UConfig::readEntry (const UCKey &k, uint &v, uint d)
{
   QVariant e = internalReadEntry (k);
   if (e.canCast (QVariant::UInt)) {
     v = e.toUInt();
     return true;
   } else {
     v = d;
     return false;
   }
}


bool UConfig::readEntry (const UCKey &k, double &v, double d)
{
   QVariant e = internalReadEntry (k);
   if (e.canCast (QVariant::Double)) {
     v = e.toDouble();
     return true;
   } else {
     v = d;
     return false;
   }
}


bool UConfig::readEntry (const UCKey &k, QFont &v, const QFont *d)
{
   QVariant e = internalReadEntry (k);
   if (e.canCast (QVariant::Font)) {
     v = e.toFont();
     return true;
   } else {
     if(d)
       v = *d;
     return false;
   }
}


bool UConfig::readEntry (const UCKey &k, QPoint &v, const QPoint *d)
{
   QVariant e = internalReadEntry (k);
   if (e.canCast (QVariant::Point)) {
     v = e.toPoint();
     return true;
   } else {
     if(d)
       v = *d;
     return false;
   }
}


bool UConfig::readEntry (const UCKey &k, QSize &v, const QSize *d)
{
   QVariant e = internalReadEntry (k);
   if (e.canCast (QVariant::Size)) {
     v = e.toSize();
     return true;
   } else {
     if(d)
       v = *d;
     return false;
   }
}


bool UConfig::readEntry (const UCKey &k, QRect &v, const QRect *d)
{
   QVariant e = internalReadEntry (k);
   if (e.canCast (QVariant::Rect)) {
     v = e.toRect();
     return true;
   } else {
     if(d)
       v = *d;
     return false;
   }
}


bool UConfig::readEntry (const UCKey &k, QColor &v, const QColor *d)
{
   QVariant e = internalReadEntry (k);
   if (e.canCast (QVariant::Color)) {
     v = e.toColor();
     return true;
   } else {
     if(d)
       v = *d;
     return false;
   }
}


bool UConfig::readEntry (const UCKey &k, QString &v, QString d)
{
   QVariant e = internalReadEntry (k);
   if (e.canCast (QVariant::String)) {
     v = e.toString();
     return true;
   } else {
     v = d;
     return false;
   }
}


bool UConfig::readEntry (const UCKey &k, QCString &v, const QCString *d)
{
   QVariant e = internalReadEntry (k);
   if (e.canCast (QVariant::CString)) {
     v = e.toCString();
     return true;
   } else {
     if(d)
       v = *d;
     return false;
   }
}


bool UConfig::deleteEntry ( const UCKey& k )
{
   UCKey::KeyType kt = k.getKeyType();
   UConfigPrivate::FAddr root = curDir->d()->getEntries( kt );
   UConfigPrivate::FAddr newRoot;
   
   switch( eCaches[ kt ]->del( root, ds, k, newRoot ) ) {
      case UConfigPrivate::Deleted:
	if ( newRoot )
	  curDir->d()->setEntries( kt, newRoot );
         return true;

      case ::UConfigPrivate::TreeEmpty:
         curDir->d()->resetEntries( kt );
         return true;

      case ::UConfigPrivate::Error:
         return false;

      default:
         qFatal( "bool UConfig::deleteEntry ( const UCKey& k ): Unknown switch." );
         return false;
      }
}


bool UConfig::deleteNode( const bool recursively )
{
  UConfigPrivate::FAddr pna;

  UConfigPrivate::Node *curNode = curDir->d();

  pna = curNode->getParentNode();

  if( !pna )  // We are root
    return false;

  if( !recursively )
    if( curNode->isInfo() )  // We have children..
      return false;

  UConfigPrivate::Node *pn;
  pn = nodes->safeGet( pna, ds );

  #ifdef __UCDEBUG
    assert( pn );
  #endif

  curDir->changeDir( pn );

  if( curNode->isLocked() ) {
    curDir->changeDir( curNode );
    return false;
  }


  UCKey::KeyType kt = UCKey::StringKey;  // FIXME
  UConfigPrivate::FAddr root = pn->getSubNodes( kt );
  UConfigPrivate::FAddr newRoot;
   
  switch( nodes->del( root, ds, *curNode->getHeader(), newRoot ) ) {
  case UConfigPrivate::Deleted:
    if ( newRoot )
      pn->setSubNodes( kt, newRoot );
    return true;
    
  case ::UConfigPrivate::TreeEmpty:
    pn->resetSubNodes( kt );
    return true;
    
  case ::UConfigPrivate::Error:
    curDir->changeDir( curNode );
    return false;
    
  default:
    qFatal( "bool UConfig::deleteNode(): Unknown switch." );
    return false;
  }
}


void UConfig::resetStatistic()
{
  nodes->resetStatistic();

  int i;
  for( i = 0; i < UCKey::knownKeyTypes; i++ )
    eCaches[ i ]->resetStatistic();

return;
}


Q_UINT32 UConfig::QStringNodesStatistic()
{
return nodes->statistic();
}


Q_UINT32 UConfig::intTreesStatistic()
{
return eCaches[ UCKey::IntegerKey ]->statistic();
}


Q_UINT32 UConfig::QStringTreesStatistic()
{
return eCaches[ UCKey::StringKey ]->statistic();
}


void UConfig::iteratorDied( UCIterator* i )
{
  #ifdef __UCDEBUG
    assert( iterators.remove( i ) );
  #else
    iterators.remove( i );
  #endif
return;
}


void UConfig::iteratorCreated( UCIterator* i )
{
  iterators.append( i );
return;
}


  // UCPathIterator calls it.
UConfigPrivate::FAddr UConfig::groupAddr( const bool sg )
{
  if( sg )
    return curDir->d()->getSubNodes(
                        UCKey::StringKey
                                   );
  else {
    UConfigPrivate::Node *parent;
    parent = nodes->safeGet( curDir->d()->getParentNode(), ds );
    if( parent )
      return parent->getSubNodes(UCKey::StringKey);
    else
      return rootPosition;
  }
}


bool UConfig::internalWriteEntry(
  const UCKey& k,
  const QVariant& e )
{
  if( e.type() == QVariant::Invalid )
    return false;

  UConfigPrivate::WEntry *we;
  UConfigPrivate::WEntry *newE;
  UCKey::KeyType kt = k.getKeyType();
  UConfigPrivate::FAddr root = curDir->d()->getEntries( kt );
  UConfigPrivate::WHeader *ph;
  UConfigPrivate::ParentType pt;

  we = eCaches[ kt ]->find( root, ds, k, &ph, &pt );

  switch( pt ) {
    case UConfigPrivate::LeftChild:
    case UConfigPrivate::RightChild:
      // Simply add new entry to exists tree.

      newE = new UConfigPrivate::WEntry(
               new UConfigPrivate::WHeader( k,
                                            lastPosition,
                                            pt,
                                            ph
                                          )
                                       );
      #ifdef __UCDEBUG
        assert( newE->update(e) );
      #else
        newE->update( e );
      #endif

      eCaches[ kt ]->add( newE, ds, &lastPosition );
      return true;


    case UConfigPrivate::FoundFromLeft:
      // update entry and update parents left pointer if need.

      if( we->update( e ) ) {
        if( we->needToRealloc() ) {
          eCaches[ kt ]->reAlloc( we, ds, &lastPosition );
          #ifdef __UCDEBUG
            qWarning( "FoundFromLeft - entry reallocating" );
          #endif
          ph->setLeft( we->getHeader() );
          ph->flush( ds );
        }
        return true;
      }
      break;


    case UConfigPrivate::FoundFromRight:
      // update entry and update parents right pointer if need.

      if( we->update( e ) ) {
        if( we->needToRealloc() ) {
          eCaches[ kt ]->reAlloc( we, ds, &lastPosition );
          #ifdef __UCDEBUG
            qWarning( "FoundFromRight - entry reallocating" );
          #endif
          ph->setRight( we->getHeader() );
          ph->flush( ds );
        }
        return true;
      }
      break;


    case UConfigPrivate::FoundFromRoot:
      // update entry and update root if we need.

      if( we->update( e ) ) {
        if( we->needToRealloc() ) {
          eCaches[ kt ]->reAlloc( we, ds, &lastPosition );

          #ifdef __UCDEBUG
            qWarning( "FoundFromRoot - entry reallocating" );
          #endif

          curDir->d()->setEntries( we );
          curDir->d()->flush( ds ); // Crash protection.
        }
        return true;
      }
      break;


    case UConfigPrivate::RootChild:
      // Create new entries tree .

      newE = new UConfigPrivate::WEntry(
               new UConfigPrivate::WHeader( k,
                                            lastPosition,
                                            pt,
                                            ph
                                          )
                                       );
      #ifdef __UCDEBUG
        assert( newE->update(e) );
      #else
        newE->update( e );
      #endif

      eCaches[ kt ]->add( newE, ds, &lastPosition );

      curDir->d()->setEntries( newE );

      if( curDir->d()->needToRealloc() )
        reAllocCurDir();

      curDir->d()->flush( ds );
      return true;

    default:
      qFatal( "bool UConfig::internalWriteEntry( ... ): unknown parent type" );
  }
return false;
}


QVariant UConfig::internalReadEntry(
  const UCKey& k, bool *ok )
{
  UConfigPrivate::WEntry *we;
  UCKey::KeyType kt = k.getKeyType();
  UConfigPrivate::FAddr root = curDir->d()->getEntries( kt );

  we = eCaches[ kt ]->find( root, ds, k );
  if( ok ) *ok  = (bool)we;
  return (we ? we->data : QVariant ());
}


UConfigPrivate::FAddr UConfig::getEntries(
  const UCKey::KeyType kt )
{
return curDir->d()->getEntries( kt );
}


void UConfig::reAllocCurDir()
{
  // We need fix all subgroups parent pointers.
  // Also parent pointer here.

  #ifdef __UCDEBUG
    qWarning("Reallocating group: %s", (const char*) curDir->d()->name().local8Bit() );
  #endif
  UConfigPrivate::FAddr lastAddr = curDir->d()->getHeader()->getPos();
  UConfigPrivate::FAddr newAddr = lastPosition;

  UConfigPrivate::FAddr pnp; // Parent node pointer

  pnp = curDir->d()->getParentNode();

  if( pnp ) { // Parent is Node

    UConfigPrivate::Node *ppn;
    UConfigPrivate::Node *pn;
    UConfigPrivate::WHeader *ph;
    UConfigPrivate::ParentType pt;

    ppn = nodes->safeGet( pnp, ds );

    UCKey::KeyType cdkt = ppn->getHeader()->getKeyType();
    pn = nodes->find( ppn->getSubNodes( cdkt ),
                      ds,
                      *(curDir->d()->getHeader()),
                      &ph,
                      &pt );

    #ifdef __UCDEBUG
      assert( pn == *curDir );
    #endif

    nodes->reAlloc( *curDir, ds, &lastPosition );

    switch( pt ) {
      case UConfigPrivate::FoundFromRoot:
        #ifdef __UCDEBUG
          assert( ppn->getSubNodes(
                    curDir->d()->getHeader()->getKeyType() ) ==
                  lastAddr );
        #endif

        ppn->setSubNodes( *curDir );

        #ifdef __UCDEBUG
          assert( !ppn->needToRealloc() );
        #endif

        ppn->flush( ds );
        break;
     case UConfigPrivate::FoundFromLeft:
        #ifdef __UCDEBUG
          assert( ph );
          assert( ph->getLeft() == lastAddr );
        #endif

        ph->setLeft( curDir->d()->getHeader() );
        ph->flush( ds );
        break;
     case UConfigPrivate::FoundFromRight:
       #ifdef __UCDEBUG
         assert( ph );
         assert( ph->getRight() == lastAddr );
       #endif

       ph->setRight( curDir->d()->getHeader() );
       ph->flush( ds );
       break;
     default:
       qFatal( "UConfig::appendGroup(..): internal DB structure error" );
    }
  }
  else {
      // Rewrite root pointer.
    nodes->reAlloc( *curDir, ds, &lastPosition );

    ds.device()->at( sizeof(dbSignature) );

    rootPosition = newAddr;
    ds << rootPosition;

    if( ds.device()->status() != IO_Ok )
      qFatal( "UConfig::appendGroup(..): Fatal datastream error" );
  }

    // Correct all subnodes for parent
  int i;
  for( i = 0; i < UCKey::knownKeyTypes; i++ ) {
    UConfigPrivate::Node *s;

    s = nodes->safeGet( curDir->d()->getSubNodes( (UCKey::KeyType)i ),
                        ds );

    if( s )
      correctParents( s, curDir->d() );
  }
return;
}


void UConfig::correctParents( UConfigPrivate::Node *n,
                              UConfigPrivate::Node *to )
{
  if( n->getHeader()->getLeft() )
    correctParents( nodes->safeGet( n->getHeader()->getLeft(),
                                    ds ),
                    to );

  n->setParent( to );
  n->flush( ds ); // Crash protection.

  if( n->getHeader()->getRight() )
    correctParents( nodes->safeGet( n->getHeader()->getRight(),
                                    ds ),
                    to );
return;
}


UCWriteStream::~UCWriteStream()
{
  if( !c->writeEntry(k, *data) )
    qFatal( "UCWriteStream::~UCWriteStream(): write failed." );

  unsetDevice();
  delete data;
}


/************************************************************
 * UCIterator                                               *
 *                                                          *
 *                                                          *
 ************************************************************/
UCIterator::UCIterator(
  UConfig *from,
  const UConfigPrivate::FAddr r )
{
  c = from;
  root = r;

  headerStack =  new UConfigPrivate::WHeaderList;

  c->iteratorCreated( this );
}


UCIterator::UCIterator( const UCIterator& from )
{
  c = from.c;
  root = from.root;

  headerStack = new UConfigPrivate::WHeaderList(*(from.headerStack));

  for( headerStack->first(); headerStack->current(); headerStack->next() )
    headerStack->current()->lock();

  c->iteratorCreated( this );
}


UCIterator::~UCIterator() 
{ 
  freeWay();  // Unlock locked headers..
  if( c ) 
    c->iteratorDied( this ); 
  delete headerStack;
} 
 
   
void UCIterator::detach()
{
  freeWay();
  c = 0;
return;
}


void UCIterator::newRoot( const UConfigPrivate::FAddr r )
{
  freeWay();
  root = r;
return;
}

UConfigPrivate::WHeader* UCIterator::pop()
{
  if( !headerStack->isEmpty() ) {
    UConfigPrivate::WHeader *h = headerStack->last();
    headerStack->remove();
    h->unLock();
  return h;
  }
  else
    return 0;
}


void UCIterator::push( UConfigPrivate::WHeader* h )
{
  h->lock();
  headerStack->append( h );
}


bool UCIterator::toLeft()
{
  freeWay();

  if( c ) { // Initialized iterator
    UConfigPrivate::WHeader *h;
    h = getHeader( root );

    if( h ) {
      while( h ) {
        push( h );
        h = getHeader( h->getLeft() );
      }

      return true;
    }
  }
return false;
}


bool UCIterator::toRight()
{
  freeWay();

  if( c ) { // Initialized iterator
    UConfigPrivate::WHeader *h;
    h = getHeader( root );

    if( h ) {
      while( h ) {
        push( h );
        h = getHeader( h->getRight() );
      }

      return true;
    }
  }
return false;
}


bool UCIterator::operator++ ()
{
  UConfigPrivate::WHeader* h = headerStack->last();

  if( !h )
    return false;

    // Try to go a bit righter.
  if( h->getRight() ) {

      // Go right and to the end of left.
    h = getHeader( h->getRight() );

      // to end of left now.
    while( h ) {
      push( h );
      h = getHeader( h->getLeft() );
    }
    return true;
  }

    // Must go up.
  UConfigPrivate::FAddr curAddr;

  pop();

  while( 1 ) {
    curAddr = h->getPos();

    h = headerStack->last();

    if( !h )
      return false; // That's end of tree.

    if( h->getLeft() == curAddr )
      return true;

    pop();
  }
}


bool UCIterator::operator--()
{
  UConfigPrivate::WHeader *h =headerStack->last();

  if( !h )
    return false;

    // Try to go a bit righter.
  if( h->getLeft() ) {

      // Go left and to the end of right.
    h = getHeader( h->getLeft() );

      // to end of right now.
    while( h ) {
      push( h );
      h = getHeader( h->getRight() );
    }
    return true;
  }

    // Must go up.
  UConfigPrivate::FAddr curAddr;

  pop();

  while( 1 ) {
    curAddr = h->getPos();

    h = headerStack->last();

    if( !h )
      return false; // That's end of tree.

    if( h->getRight() == curAddr )
      return true;

    pop();
  }
}

inline bool UCIterator::isValid() const
{ 
  return c && root && headerStack->current(); 
}
UConfigPrivate::WHeader* UCIterator::top() 
{ 
  return headerStack->last(); 
}


#ifdef __UCDEBUG
void UCIterator::debugWay()
{
  for( headerStack->first(); headerStack->current(); headerStack->next() ) {
    qDebug( "Key: " );
    switch( headerStack->current()->getKeyType() ) {
      case UCKey::StringKey:
        qDebug( "%s", (const char *)headerStack->current()->getStringKey().local8Bit() );
        break;
      case UCKey::IntegerKey:
        qDebug( "%s", (const char *)headerStack->current()->getIntegerKey() );
        break;
      default:
        qDebug( "Unknown key type" );
    }
  }
return;
}
#endif


/************************************************************
 * UCPathIterator                                           *
 *                                                          *
 *                                                          *
 ************************************************************/

UCPathIterator::UCPathIterator( UConfig *from,  const bool subGroupIterator )
  :
  UCIterator( from, from->groupAddr(subGroupIterator) )
{
  toLeft();
}


UCPathIterator::UCPathIterator( const UCPathIterator& from )
  :
  UCIterator( from )
{
}


  // FIXME! optimize not to copy stack
UCPathIterator::UCPathIterator(
  const UCPathIterator& from,
  const bool subGroupIterator )
  :
  UCIterator( from )
{
  if( subGroupIterator ) {
    if( from.isValid() ) {
      UConfigPrivate::Node* n;
      n = c->nodes->safeGet( top(),
                             c->ds );

      #ifdef __UCDEBUG
        assert( n );
      #endif

      newRoot( n->getSubNodes( UCKey::StringKey ) );
      toLeft();
    }
    else
      qFatal( "Can not construct subgroup iterator from invalid iterator" );
  }
}



UCPathIterator::~UCPathIterator()
{
}


UConfigPrivate::WHeader* UCPathIterator::getHeader( const UConfigPrivate::FAddr from )
{
  if( c )
    return c->nodes->safeGetHeader( from, c->ds );
  else
    return 0;
}



QString UCPathIterator::group()
{
   if( top() )
      return top()->getStringKey();

return QString::null;
}


QString UCPathIterator::path()
{
  if( top() ) {
    UConfigPrivate::Node * n;
    n = c->nodes->safeGet( top(), c->ds );

    return c->path( n );
  }

return QString::null;
}


/************************************************************
 * UCEntriesIterator                                        *
 *                                                          *
 *                                                          *
 ************************************************************/

UCEntriesIterator::UCEntriesIterator(
  UConfig *c,
  const int ) : UCIterator( c,
            c->getEntries(UCKey::IntegerKey) )
{
  kt = UCKey::IntegerKey;
  toLeft();
}


UCEntriesIterator::UCEntriesIterator(
  UConfig *c,
  const QString& ) : UCIterator( c,
            c->getEntries(UCKey::StringKey) )
{
  kt = UCKey::StringKey;
  toLeft();
}


UCEntriesIterator::UCEntriesIterator(
  const UCEntriesIterator& from
)
  : UCIterator( from )
{
  kt = from.kt;
}


UCEntriesIterator::~UCEntriesIterator()
{
}


UConfigPrivate::WHeader* UCEntriesIterator::getHeader( UConfigPrivate::FAddr a )
{
  if( c )
    return c->eCaches[ kt ]->safeGetHeader( a, c->ds );
  else
    return 0;
}


QVariant UCEntriesIterator::readEntry( bool *ok )
{
  UConfigPrivate::WEntry *we = 0;

  if( top() )
    we = c->eCaches[ kt ]->safeGet( top(), c->ds );

  if( ok ) *ok = (bool)we;
  return (we ? we->data : QVariant());
}
 

 QVariant UCEntriesIterator::getKey()
 { 
  if( isValid() ) 
    return top()->getKey(); 
//    return ((WHeader *)top())->getKey();
 return QVariant(); 
} 
