/***************************************************************************
    description          : corewidget.cpp
    begin                : Tue Apr 25 2000
    copyright            : (C) 2000 by Florian Veskov
    email                : napalmbox@icqmail.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "corewidget.h"
#include <qkeycode.h>
#include <stdlib.h>
#include <qevent.h>
#include <qpainter.h>
#include <qpixmap.h>
#include <qframe.h>
#include <qmlined.h>
#include "corewidget.moc"

RestrictedLine::RestrictedLine(QWidget *parent, const char *name, const char *valid)
  : QLineEdit(parent, name)
{
  qsValidChars = valid;
  return;
}

RestrictedLine::~RestrictedLine(void)
{
  return;
}

void RestrictedLine::keyPressEvent(QKeyEvent *e)
{
  if (e->key() == Key_Enter || e->key() == Key_Return)
  {
    emit returnPressed();
    return;
  }
  if (e->ascii() < 32)
  {
    QLineEdit::keyPressEvent(e);
    return;
  }
  if (!qsValidChars.isEmpty() && !qsValidChars.contains((char)e->ascii()))
  {
    emit (invalidChar(e->key()));
    return;
  }
  else
    QLineEdit::keyPressEvent(e);
  return;
}


void RestrictedLine::setValidChars(const char *valid)
{
  qsValidChars = valid;
  return;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////

KIntegerLine::KIntegerLine(QWidget *parent, const char *name, EditLineType t)
  : RestrictedLine(parent, name, "0123456789")
{
  switch (lineType = t)
  {
    case EditTypeOct:
      setValidChars("01234567");
      break;
    case EditTypeHex:
      setValidChars("0123456789ABCDEFabcdef");
      break;
    default:
      break;
  }
  connect(this, SIGNAL(textChanged(const QString&)),
          this, SLOT(internalValueChanged()));
}

KIntegerLine::KIntegerLine(void) : RestrictedLine()
{
  return;
}

KIntegerLine::~KIntegerLine(void)
{
  return;
}

EditLineType KIntegerLine::getType(void)
{
  return lineType;
}


int KIntegerLine::value(void)
{
  QString s = text();
  return getValue(s);
}


void KIntegerLine::setValue(int value)
{
  QString s;
  putValue(s, value);
}


void KIntegerLine::keyPressEvent(QKeyEvent *e)
{
  QString tmp(text());
  int i;

  switch (e->key())
  {
    case Key_Up:
      putValue(tmp, getValue(tmp)+1);
      break;
    case Key_Down:
      i = getValue(tmp);
      if (i>0) i--;
      putValue(tmp, i);
      break;
    case Key_Prior:
      incValue(tmp, getValue(tmp));
      break;
    case Key_Next:
      decValue(tmp, getValue(tmp));
      break;
    default: // switch (e->key())
      RestrictedLine::keyPressEvent(e);
      break;
  }
  return;
}


int KIntegerLine::getValue(QString &s)
{
  return (strtol(s.latin1(), 0, (int)lineType));
}

void KIntegerLine::putValue(QString &s, int val)
{
  switch (lineType)
  {
    case EditTypeOct:
      s.sprintf("%o", val);
      break;
    case EditTypeDec:
      s.sprintf("%d", val);
      break;
    case EditTypeHex:
      s.sprintf("%X", val);
      break;
  }
  setText(s);
  return;
}

void KIntegerLine::incValue(QString &s, int val)
{
  switch (lineType)
  {
    case EditTypeOct:
      val += 8;
      s.sprintf("%o", val);
      break;
    case EditTypeDec:
      val += 10;
      s.sprintf("%d", val);
      break;
    case EditTypeHex:
      val +=16;
      s.sprintf("%X", val);
      break;
  }
  setText(s);
  return;
}

void KIntegerLine::decValue(QString &s, int val)
{
  switch (lineType)
  {
    case EditTypeOct:
      if (val > 7)
        val -= 8;
      s.sprintf("%o", val);
      break;
    case EditTypeDec:
      if (val > 9)
        val -= 10;
      s.sprintf("%d", val);
      break;
    case EditTypeHex:
      if (val > 15)
        val -= 16;
      s.sprintf("%X", val);
      break;
  }
  setText(s);
  return;
}


void KIntegerLine::internalValueChanged(void)
{
  emit valueChanged(value());
  return;
}

ULine::ULine(QWidget *parent, bool readonly) : QLineEdit(parent)
{
  baseRO = palette().disabled().base();
  baseRW = palette().inactive().base();
  SetReadOnly(readonly);
  return;
}
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void ULine::SetReadOnly(bool b)
{
  m_bReadOnly = b;

  QPalette pal;
  QColorGroup cg;
  cg.setColor(QColorGroup::Foreground, black);
  cg.setColor(QColorGroup::Button, QColor(228, 228, 228));
  cg.setColor(QColorGroup::Light, white);
  cg.setColor(QColorGroup::Midlight, QColor(241, 241, 241));
  cg.setColor(QColorGroup::Dark, QColor(114, 114, 114));
  cg.setColor(QColorGroup::Mid, QColor(152, 152, 152));
  cg.setColor(QColorGroup::Text, black);
  cg.setColor(QColorGroup::BrightText, white);
  cg.setColor(QColorGroup::ButtonText, black);
  cg.setColor(QColorGroup::Base, QColor(228, 228, 228));
  cg.setColor(QColorGroup::Background, QColor(228, 228, 228));
  cg.setColor(QColorGroup::Shadow, black);
  cg.setColor(QColorGroup::Highlight, black);
  cg.setColor(QColorGroup::HighlightedText, white);
  pal.setActive(cg);
  cg.setColor(QColorGroup::Foreground, black);
  cg.setColor(QColorGroup::Button, QColor(228, 228, 228));
  cg.setColor(QColorGroup::Light, white);
  cg.setColor(QColorGroup::Midlight, white);
  cg.setColor(QColorGroup::Dark, QColor(114, 114, 114));
  cg.setColor(QColorGroup::Mid, QColor(152, 152, 152));
  cg.setColor(QColorGroup::Text, black);
  cg.setColor(QColorGroup::BrightText, white);
  cg.setColor(QColorGroup::ButtonText, black);
  cg.setColor(QColorGroup::Base, QColor(228, 228, 228));
  cg.setColor(QColorGroup::Background, QColor(228, 228, 228));
  cg.setColor(QColorGroup::Shadow, black);
  cg.setColor(QColorGroup::Highlight, black);
  cg.setColor(QColorGroup::HighlightedText, white);
  pal.setInactive(cg);
  cg.setColor(QColorGroup::Foreground, QColor(128, 128, 128));
  cg.setColor(QColorGroup::Button, QColor(228, 228, 228));
  cg.setColor(QColorGroup::Light, white);
  cg.setColor(QColorGroup::Midlight, white);
  cg.setColor(QColorGroup::Dark, QColor(114, 114, 114));
  cg.setColor(QColorGroup::Mid, QColor(152, 152, 152));
  cg.setColor(QColorGroup::Text, black);
  cg.setColor(QColorGroup::BrightText, white);
  cg.setColor(QColorGroup::ButtonText, QColor(128, 128, 128));
  cg.setColor(QColorGroup::Base, QColor(228, 228, 228));
  cg.setColor(QColorGroup::Background, QColor(228, 228, 228));
  cg.setColor(QColorGroup::Shadow, black);
  cg.setColor(QColorGroup::Highlight, black);
  cg.setColor(QColorGroup::HighlightedText, white);
  pal.setDisabled(cg);
  setPalette(pal);
  return;
}

void ULine::setEnabled(bool _b)
{
  QLineEdit::setEnabled(_b);
  return;
}
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
MLine::MLine(QWidget *parent, bool readonly) : QMultiLineEdit(parent)
{
  setWordWrap(WidgetWidth);
  setWrapPolicy(AtWhiteSpace);
  baseRO = palette().disabled().base();
  baseRW = palette().inactive().base();
  SetReadOnly(readonly);
  return;
}
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void MLine::SetReadOnly(bool b)
{
  m_bReadOnly = b;

  QPalette pal;
  QColorGroup cg;
  cg.setColor(QColorGroup::Foreground, black);
  cg.setColor(QColorGroup::Button, QColor(228, 228, 228));
  cg.setColor(QColorGroup::Light, white);
  cg.setColor(QColorGroup::Midlight, QColor(241, 241, 241));
  cg.setColor(QColorGroup::Dark, QColor(114, 114, 114));
  cg.setColor(QColorGroup::Mid, QColor(152, 152, 152));
  cg.setColor(QColorGroup::Text, black);
  cg.setColor(QColorGroup::BrightText, white);
  cg.setColor(QColorGroup::ButtonText, black);
  cg.setColor(QColorGroup::Base, QColor(228, 228, 228));
  cg.setColor(QColorGroup::Background, QColor(228, 228, 228));
  cg.setColor(QColorGroup::Shadow, black);
  cg.setColor(QColorGroup::Highlight, black);
  cg.setColor(QColorGroup::HighlightedText, white);
  pal.setActive(cg);
  cg.setColor(QColorGroup::Foreground, black);
  cg.setColor(QColorGroup::Button, QColor(228, 228, 228));
  cg.setColor(QColorGroup::Light, white);
  cg.setColor(QColorGroup::Midlight, white);
  cg.setColor(QColorGroup::Dark, QColor(114, 114, 114));
  cg.setColor(QColorGroup::Mid, QColor(152, 152, 152));
  cg.setColor(QColorGroup::Text, black);
  cg.setColor(QColorGroup::BrightText, white);
  cg.setColor(QColorGroup::ButtonText, black);
  cg.setColor(QColorGroup::Base, QColor(228, 228, 228));
  cg.setColor(QColorGroup::Background, QColor(228, 228, 228));
  cg.setColor(QColorGroup::Shadow, black);
  cg.setColor(QColorGroup::Highlight, black);
  cg.setColor(QColorGroup::HighlightedText, white);
  pal.setInactive(cg);
  cg.setColor(QColorGroup::Foreground, QColor(128, 128, 128));
  cg.setColor(QColorGroup::Button, QColor(228, 228, 228));
  cg.setColor(QColorGroup::Light, white);
  cg.setColor(QColorGroup::Midlight, white);
  cg.setColor(QColorGroup::Dark, QColor(114, 114, 114));
  cg.setColor(QColorGroup::Mid, QColor(152, 152, 152));
  cg.setColor(QColorGroup::Text, black);
  cg.setColor(QColorGroup::BrightText, white);
  cg.setColor(QColorGroup::ButtonText, QColor(128, 128, 128));
  cg.setColor(QColorGroup::Base, QColor(228, 228, 228));
  cg.setColor(QColorGroup::Background, QColor(228, 228, 228));
  cg.setColor(QColorGroup::Shadow, black);
  cg.setColor(QColorGroup::Highlight, black);
  cg.setColor(QColorGroup::HighlightedText, white);
  pal.setDisabled(cg);
  setPalette(pal);
  return;
}

void MLine::setEnabled(bool _b)
{
  QMultiLineEdit::setEnabled(_b);
  return;
}


/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
ITButton::ITButton(QWidget *parent, const char *name) : QPushButton(parent, name)
{
  return;
}

ITButton::~ITButton(void)
{
  return;
}
void ITButton::setIT(const QPixmap &itpix, const QString &ittext)
{
  bpix = itpix;
  btext= ittext;
  return;
}
void ITButton::drawButtonLabel(QPainter *p)
{
  p->drawPixmap(6, 4, bpix);
  p->drawText(bpix.width()+13, 0, width()-13-bpix.width(), height(), AlignLeft|AlignVCenter, btext);
  return;
}
